package tsp;

import java.util.*;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.graphics.*;

/**
 * This class implements the actual MST-TSP algorithm
 * In short, the algorithm works by clustering cities by running the first iterations
 * of Boruvka's MST algorithm to get groups of cities. This groups are joined as long
 * as their size is below a certain (configurable) threshold.
 * Each of these groups or clusters is then replaced by a single representative
 * meta-city. The algorithm continues on the next level clustering the meta-cities
 * until we have only one meta-city left.
 * Then we enter the refining phase of the algorithm which expands the meta-cities by
 * replacing it with its children (the cities which originally got collapsed into this
 * meta-city). The cities are put into a tour which starts out being of zero length
 * and containing only the topmost meta-city. During the refining phase the tour grows
 * longer. It is important for the algorithm to determine a suitable order of the 
 * cities in which they are inserted into the tour. In the current implementation this
 * is achieved by finding the optimal permutation of the children-cities which results
 * in the shortest possible order of the cities within one cluster.
 * 
 * @author Andreas Lehrbaum
 */
public class MSTPyramid implements Runnable {
	Vector<TSPInstance> tsps;

	Hashtable<City, Vector<City>> children;
	Tour tour = null;

	public void run() {
		cluster();
		refine();
	}
	
	// this parameter limits the size of the individual clusters at each level and has
	// great impact on the resulting pyramid structure
	int maxClusterSize = 7;

	/**
	 * This method creates a new pyramid suitable for running the MST-TSP algorithm 
	 * @param tsp The TSP instance this pyramid should represent
	 */
	public MSTPyramid(TSPInstance tsp) {
		tsps = new Vector<TSPInstance>();
		tsps.add(tsp);
		children = new Hashtable<City, Vector<City>>();
	}

	/**
	 * This method initiates the clustering of the cities
	 */
	public void cluster() {
		if (tsps.get(0).cities.size() <= 1) return;
		int level = 1;
		
		cluster(0);
		// cluster as long as there are cities still to cluster
		while (tsps.get(level).cities.size() > 1) {
			// Running first iteration of Boruvka's MST-algorithm
			cluster(level++);
		}
	}
	
	/**
	 * This method does the actual clustering at a given level 
	 * @param level The level at which to cluster the cities, starting at 0 (bottom)
	 */
	public void cluster(int level) {
		// for the Boruvka-algorithm we need a set of trees which we can join later
		// therefore we use a Hashtable to store the treeID of each City
		Hashtable treeId = new Hashtable<City, Integer>();
		Vector<City> cities = tsps.get(level).cities;
		
		// start with each city in its own tree
		for (int i = 0; i < cities.size(); i++) {
			treeId.put(cities.get(i), new Integer(i));
		}

		int size = cities.size();
		int numOfTrees = size;

		// for each city: find the closest city and join the two trees
		for (int i = 0; i < cities.size(); i++) {
			City c = cities.get(i);
			City cClosest = c.getClosestCity();
			
			// join the trees of c and cClosest if they are in different trees
			if (!treeId.get(c).equals(treeId.get(cClosest))) {
				// check if the resulting tree would exceed maxClusterSize -> in this case we do not join them!
				int combinedSize = 0;
				for (Enumeration e = treeId.keys(); e.hasMoreElements(); ) {
					Integer id = (Integer)treeId.get((City)e.nextElement());
					if (id.equals(treeId.get(c)) || id.equals(treeId.get(cClosest))) {
						if (++combinedSize > maxClusterSize)
							break;
					}
				}
				if (combinedSize <= maxClusterSize) {
					numOfTrees--;
					// change all treeIds equal to the one of cClosest to the treeId of c
					Integer oldTreeId = (Integer) treeId.get(cClosest);
					for (int i2 = 0; i2 < cities.size(); i2++) {
						City c2 = cities.get(i2);
	
						if (((Integer) treeId.get(c2)).equals(oldTreeId)) {
							treeId.put(c2, treeId.get(c));
						}
					}
				}
			}
		}

		// create new TSP for the next level
		// first gather the clusters in a Hashtable, accessed by their treeId
		Hashtable<Integer, Vector<City>> clusters = new Hashtable<Integer, Vector<City>>();
		for (Enumeration e = cities.elements(); e.hasMoreElements();) {
			City c = (City) e.nextElement();
			Vector<City> cluster = clusters.get(treeId.get(c));
			if (cluster == null) {
				cluster = new Vector<City>();
				clusters.put((Integer) treeId.get(c), cluster);
			}
			cluster.add(c);
		}

		// now create a new TSP where each cluster forms a city with the cities
		// of the cluster as its children
		TSPInstance nextTSP = new TSPInstance();
		for (Enumeration e = clusters.keys(); e.hasMoreElements();) {
			Integer key = (Integer) e.nextElement();
			nextTSP.addCity(createMetaCity(clusters.get(key)));
		}

		// insert the new TSP
		tsps.insertElementAt(nextTSP, level + 1);
	}

	/**
	 * This method collapses a cluster of cities into one meta-city positioned at the center of the cluster
	 * @param cities The vector of the cities to cluster
	 * @return The newly created meta-city
	 */
	public City createMetaCity(Vector<City> cities) {
		int x = 0, y = 0;
		
		// calculate the average of the positions
		// TODO: parametrize!!!
		for (Enumeration e = cities.elements(); e.hasMoreElements();) {
			City c = (City) e.nextElement();
			x += c.x;
			y += c.y;
		}
		x /= cities.size();
		y /= cities.size();

		// create the new city
		City newCity = new City(x, y);
		// remember the children of this city
		children.put(newCity, cities);

		return newCity;
	}

	/**
	 * This method initiates the refining-phase of the algorithm
	 *
	 */
	public void refine() {
		tour = new Tour(null);
		int level = tsps.size() - 1;

		// build trivial starting-tour
		// TODO: this could be improved in case we don't have only one single city on top
		for (Iterator i = tsps.get(level).cities.iterator(); i.hasNext();) {
			tour.cities.add((City) i.next());
		}
		refine(tour);
	}

	/**
	 * This method calls refine for each city in the current tour
	 * @param tour The tour consisting of the cities refined so far
	 */
	public void refine(Tour tour) {
		// loop over all cities of the tour and refine them one by one
		// TODO: this order could (should) be parametrized
		for (int i = 0; i < tour.cities.size(); i++) {
			City c = tour.cities.get(i);
			refine(tour, c);
		}
	}

	/**
	 * This method does the actual refining of one city
	 * @param tour The tour consisting of the cities refined so far
	 * @param city The city we want to expand
	 */
	public void refine(Tour tour, City city) {
		if (children.get(city) == null)
			return;

		int[] permutation = new int[children.get(city).size()];
		int[] bestPermutation = new int[permutation.length];
		double bestLength = Double.MAX_VALUE;
		double length;
		
		// initiate the permutation-array
		for (int i = 0; i < permutation.length; i++) {
			permutation[i] = i;
		}
		
		int pos = tour.cities.indexOf(city);
		if (pos == -1)
			return;

		// first remove the city
		tour.cities.remove(pos);

		// then replace it by its children, evaluating every permutation
		int fakt = fakt(permutation.length);
		for (int i = 0; i < fakt; i++) {
			for (int j = 0; j < permutation.length; j++) {
				tour.cities.insertElementAt(children.get(city).get(
						permutation[j]), pos);
			}

			length = tour.getLength();
			if (length < bestLength) {
				bestLength = length;
				// remember the best permutation
				for (int j = 0; j < permutation.length; j++) {
					bestPermutation[j] = permutation[j];
				}
			}
			// then remove the children again
			for (int j = 0; j < permutation.length; j++) {
				tour.cities.remove(pos);
			}

			getNextPermutation(permutation);
		}

		// here we have found the best permutation -> use it
		for (int i = 0; i < permutation.length; i++) {
			tour.cities.insertElementAt(children.get(city).get(
					bestPermutation[i]), pos);
		}
		
		// go down one step, refine at the next lower level
		for (int i = 0; i < children.get(city).size(); i++) {
			refine(tour, children.get(city).get(i));
		}
	}

	/**
	 * Caclulate the faculty
	 */
	public int fakt(int x) {
		int result = x;
		while (x > 1) {
			x--;
			result *= x;
		}
		return result;
	}

	/**
	 * This method if called repeatedly returns all permutations of a given array
	 * @param field The array to permutate
	 */
	public void getNextPermutation(int[] field) {
		int i = field.length - 1;
		int j = field.length;

		try {
			while (field[i - 1] >= field[i])
				i--;
			while (field[j - 1] <= field[i - 1])
				j--;
		} catch (ArrayIndexOutOfBoundsException e) {
			return;
		}

		// swap values at positions (i-1) and (j-1)
		int x = field[i - 1];
		field[i - 1] = field[j - 1];
		field[j - 1] = x;

		i++;
		j = field.length;

		while (i < j) {
			x = field[i - 1];
			field[i - 1] = field[j - 1];
			field[j - 1] = x;
			i++;
			j--;
		}
	}
}
